### MP Fénelon Sainte-Marie --- Informatique
### Chapitre 07 --- Codes correcteurs d'erreur
##############################################################################

import numpy.random as rd
import time

##############################################

def AddBinaire(a,b):
    if a==b:
        return 0
    else:
        return 1

##############################################
# A partir d'un paquet (=liste) de 4 données,
# construit un code à 7 bits (4 de données, 3 de correction)

def Encodage(d):
    c = [0 for i in range(7)]
    # les quatre bits de données :
    c[2] = d[0]
    c[4] = d[1]
    c[5] = d[2]
    c[6] = d[3]
    # les trois bits de contrôle :
    c[0] = AddBinaire(d[0],AddBinaire(d[1],d[3]))
    c[1] = AddBinaire(d[0],AddBinaire(d[2],d[3]))
    c[3] = AddBinaire(d[1],AddBinaire(d[2],d[3]))
    
    return c

print("********** construction des données et du code associé **********")
donnees = [1,0,1,0]
code = Encodage(donnees)
#print("données                      :",donnees)
#print("code associé                 :",code)

##############################################
# A partir d'un code (4 bits de données et 3 bits de correction)
# reconstruction du paquet de données initial (à 4 bits)

def Decodage(c):
    d = [0 for i in range(4)]
    # les quatre bits de données :
    d[0] = c[2]
    d[1] = c[4]
    d[2] = c[5]
    d[3] = c[6]
    return d

#decode = Decodage(code)
#print("données initiales :",decode)

##############################################
# Vérification si un code à 7 bits est correct

def Verification(c):
    if c[0] != AddBinaire(c[2],AddBinaire(c[4],c[6])):
        return False
    if c[1] != AddBinaire(c[2],AddBinaire(c[5],c[6])):
        return False
    if c[3] != AddBinaire(c[4],AddBinaire(c[5],c[6])):
        return False
    return True

#print("vérification",Verification(code))
    
##############################################
# Correction d'un code cc à 7 bits avec au plus une erreur,
# renvoyant le code c corrigé

def Corrige(cc):
    c = list(cc)
    verif0 = ( c[0] == AddBinaire(c[2],AddBinaire(c[4],c[6])) )
    verif1 = ( c[1] == AddBinaire(c[2],AddBinaire(c[5],c[6])) )
    verif2 = ( c[3] == AddBinaire(c[4],AddBinaire(c[5],c[6])) )
    if verif0:
        # erreur pas dans disque0
        if verif1:
            # erreur pas dans disque1
            # alors erreur en p2
            c[3] = AddBinaire(c[3],1)
        else:
            # erreur dans disque1
            # donc erreur en p1 ou d2
            if verif2:
                # erreur pas dans disque 2
                # alors erreur en p1
                c[1] = AddBinaire(c[1],1)
            else:
                # erreur dans disque 2
                # alors erreur en d2
                c[5] = AddBinaire(c[5],1)
    else:
        # ALORS l'erreur est dans disque0
        if verif1:
            # erreur pas dans disque1
            # alors erreur en d1 ou p0
            if verif2:
                # erreur pas dans disque 2
                # alors erreur en p0
                c[0] = AddBinaire(c[0],1)
            else:
                # erreur dans le disque 2
                # alors erreur en d1
                c[4] = AddBinaire(c[4],1)
        else:
            # erreur dans disque1
            # donc erreur en d0 ou d3
            if verif2:
                # erreur pas dans disque2
                # alors erreur en d0
                c[2] = AddBinaire(c[2],1)
            else:
                # erreur dans disque2
                # alors erreur en d3
                c[6] = AddBinaire(c[6],1)
    return c

print("********** test manuel de la fonction de bruitage **********")

codebruite = list(code)
codebruite[2] = AddBinaire(codebruite[2],1)
#print("code bruité à la main        :",codebruite)
codecorrige = Corrige(codebruite)
#print("code bruité corrigé          :",codecorrige)
#print("données initiales récupérées :",Decodage(codecorrige))

##############################################
# Bruitage aléatoire !

# Construction d'un code new_c copie du code c et possédant,
# avec une probabilité epsilon, un bit (sur sept) modifié par rapport à c.

def Bruitage(c,epsilon):
    new_c = list(c)
    if rd.random()<epsilon:
        # choix au hasard d'un indice
        i = rd.randint(0,7)
        new_c[i] = AddBinaire(new_c[i],1)
    return new_c

"""
print("********** test intensif de bruitage et correction **********")
N = 10000
eps = 0.0001
print("nombre de tests:",N)
print("taux d'erreur de transmission :",eps)
temps = time.time()
temps_encodage = 0
temps_correction = 0
temps_decodage = 0

for n in range(N):
    # création d'une liste de données alétoire
    d = list(rd.randint(0, 2, 4))
    
    # encodage
    t_e = time.time()
    c = Encodage(d)
    temps_encodage = temps_encodage + time.time() - t_e
    
    # bruitage
    b = Bruitage(c,eps)
    
    # correction
    t_c = time.time()
    corr = Corrige(b)
    temps_correction = temps_correction + time.time() - t_c
    
    # décodage
    t_d = time.time()
    dd = Decodage(corr)
    temps_decodage = temps_decodage + time.time() - t_d
    if c != b:
        print(n,"code:",c,"bruitage:",b,"correction:",corr)
print("temps d'encodage            :",temps_encodage)
print("temps de correction         :",temps_correction)
print("temps de décodage           :",temps_decodage)
print("encodage+correction+décodage:",temps_encodage+temps_correction+temps_decodage)
print("temps total                 :",time.time()-temps)
"""

##############################################
# Multi bruitage aléatoire !

# Construction d'un code new_c copie du code c et tel que chacun de ses sept
# bits est modifié avec une probabilité epsilon.

def MultiBruitage(c,epsilon):
    new_c = list(c)
    nb_erreurs = 0
    for i in range(7):
        if rd.random()<epsilon:
            nb_erreurs = nb_erreurs +1
            new_c[i] = AddBinaire(new_c[i],1)
    if nb_erreurs > 1:
        print("********** au moins deux erreurs **********")
        print("initial :",c)
        print("bruité  :",new_c)
    return new_c

"""
print("********** test de la fonction de multi-bruitage **********")
N = 1000
n = 0
eps = 0.0001
print("sur un total de",N,"codes multi-bruités avérés avec P(erreur)=",eps,",")
print("combien de codes avec au moins 2 erreurs ?")
while n<N:
    codebruite2 = MultiBruitage(code,eps)
    if codebruite2 != code:
        n = n+1
"""

print("********** test intensif de multi-bruitage et correction **********")
N = 18000000 * 4 // 7
#N = 18000000 * 4 // 7   # 3 minutes de calcul
eps = 10**(-7)
print("examen de",N,"paquets de données de 4 bits chacun")
print("soit au total",N*7//4,"bits (données+contrôle) traités")
print("P(erreur de transmission)=",eps)
temps = time.time()
temps_creation = 0
temps_encodage = 0
temps_bruitage = 0
temps_correction = 0
temps_decodage = 0
r = N//100
NbErreurs = 0
AuMoinsDeuxErreurs = 0
for n in range(N):
    if n%r == 0:
        print(n//r,"% effectués")
    # création d'une liste de données alétoire
    t_crea = time.time()
    d = list(rd.randint(0, 2, 4))
    temps_creation = temps_creation + time.time() - t_crea
    
    # encodage
    t_e = time.time()
    c = Encodage(d)
    temps_encodage = temps_encodage + time.time() - t_e
    
    # bruitage
    t_b = time.time()
    b = MultiBruitage(c,eps)
    temps_bruitage = temps_bruitage + time.time() - t_b
    
    # correction, uniquement si nécessaire
    t_c = time.time()
    if Verification(b)==False:
        corr = Corrige(b)
        NbErreurs = NbErreurs + 1
    else:
        corr = list(b)
    temps_correction = temps_correction + time.time() - t_c
    
    # décodage
    t_d = time.time()
    dd = Decodage(corr)
    temps_decodage = temps_decodage + time.time() - t_d
    
    if b!=c:
        print('Multibruitage :')
        print("d avant         ",d,"c avant bruitage",c)
        print("c après bruitage",b,"d après décodage",dd)
    if dd != d:
        AuMoinsDeuxErreurs = AuMoinsDeuxErreurs + 1
        print("au moins deux erreurs !")
        print("d avant",d)
        print("d après",dd)

print("temps de création des données :",temps_creation)
print("temps d'encodage              :",temps_encodage)
print("temps de bruitage             :",temps_bruitage)
print("temps de correction           :",temps_correction)
print("temps de décodage             :",temps_decodage)
print("encod + bruit + corr + décod  :",temps_encodage+temps_bruitage+temps_correction+temps_decodage)
print("temps total                   :",time.time()-temps)
print(NbErreurs,"paquet(s) avec 1 erreur détectée & corrigée sur",N,"paquets")
print(AuMoinsDeuxErreurs,"paquet(s) avec 2 erreurs")